use crate::packet::identifier::Identifier;
use crate::prelude::{as_u32_be, as_u64_be};

#[derive(Debug, Eq, PartialEq)]
pub enum DataType {
    // Basic types
    Nothing,
    Switch(bool),
    Slider(u8),
    Text(String),

    // Cosmetic
    RGB(u8, u8, u8),

    // Time & Space
    Seconds(u64),

    // Domo Types
    NodeRef(Identifier),
}

impl DataType {
    pub fn get_data_size(data_type: u8) -> usize {
        match data_type {
            0x01 => 1,
            0x02 => 1,
            0x03 => 256,
            0x10 => 3,
            0x90 => 8,
            0xF0 => 4,
            _ => 0,
        }
    }

    pub fn get_data(buf: Vec<u8>) -> Vec<DataType> {
        let mut data = vec![];
        let mut pieces = vec![];
        let mut byte = buf.iter().peekable();

        while let Some(&c) = byte.next() {
            let mut piece = vec![c];
            let mut len = DataType::get_data_size(c);
            while len > 0 {
                if let Some(_) = byte.peek() {
                    piece.push(*byte.next().unwrap());
                }
                len -= 1;
            }
            if piece.len() == DataType::get_data_size(c) + 1 {
                pieces.push(piece);
            } else {
                pieces.push(vec![0x00]);
            }
        }

        for piece in pieces {
            data.push(DataType::from(piece))
        }

        data
    }
}

macro_rules! impl_data_type {
    ($v: ident, $len: literal, $($t:tt)*) => {
        if $v.len() >= $len {
            $($t)*
        } else {
            DataType::Nothing
        }
    };
}

impl From<Vec<u8>> for DataType {
    /// Turn a Vec<u8> into a single data type.
    /// Returns Datatype::Nothing if it's nothing or can find nothing.
    fn from(value: Vec<u8>) -> Self {
        if value.len() == 0 {
            return DataType::Nothing;
        };
        match value[0] {
            0x01 => impl_data_type!(value, 1, DataType::Switch(value[1] != 0)),
            0x02 => impl_data_type!(value, 1, DataType::Slider(value[1])),
            0x03 => impl_data_type!(
                value,
                256,
                if let Ok(s) = String::from_utf8(value[1..257].to_vec()) {
                    DataType::Text(s)
                } else {
                    DataType::Nothing
                }
            ),
            0x10 => impl_data_type!(value, 3, DataType::RGB(value[1], value[2], value[3])),
            0x90 => impl_data_type!(value, 8, DataType::Seconds(as_u64_be(&value[1..9]))),
            0xF0 => impl_data_type!(value, 4, DataType::NodeRef(Identifier::from(&value[1..5]))),
            _ => DataType::Nothing,
        }
    }
}

impl Into<Vec<u8>> for DataType {
    fn into(self) -> Vec<u8> {
        match self {
            DataType::Nothing => vec![0x00],
            DataType::Switch(b) => vec![0x01, b as u8],
            DataType::Slider(v) => vec![0x02, v],
            DataType::Text(s) => {
                let mut bytes = vec![0x03];
                bytes.extend(s.into_bytes());
                bytes
            },
            DataType::RGB(r, g, b) => vec![0x10, r, g, b],
            DataType::Seconds(s) => {
                let mut bytes = vec![0x90];
                bytes.extend_from_slice(s.to_be_bytes().as_ref());
                bytes
            },
            DataType::NodeRef(id) => {
                let mut bytes = vec![0x90];
                bytes.extend_from_slice(id.bytes.as_ref());
                bytes
            },
        }
    }
}